/*
 * Created on 05.01.2006
 *
 * Copyright --- All rights reserved.
 * ITI, Universität Karlsruhe, Germany.
 */
package de.uka.algo.GT3D;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.media.j3d.Appearance;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.DirectionalLight;
import javax.media.j3d.Group;
import javax.media.j3d.Material;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.AxisAngle4d;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.geometry.Cone;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.Sphere;

import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.Vertex;

public class BFSVisualizer implements Scene3DProvider {

//	boolean restrictArrowCones = true;

	
	/**
	 * A reference on the underlying graph which is drawn.
	 */
	private DirectedGraph theGraph = null;
	
	/** The root of it all */
	private BranchGroup sceneRoot = null;
	/** The root of everything that is going to be dynamic */
	private BranchGroup dynamicRoot = null; 
	/** The root for everything static */
	private BranchGroup staticRoot = null;
	/** The root for all the nodes */
	private BranchGroup nodeRoot = null;
	/** The root for all the edges */
	private BranchGroup edgeRoot = null;
	/** The root for all the lights */
	private BranchGroup dynLightRoot = null; 
	
	// Maps for looking up stuff about nodes and edges
	private Map<Vertex, Vector3f> nodeToLocation = new HashMap<Vertex, Vector3f>();
	private Map<Vertex, Sphere> nodeToSphere = new HashMap<Vertex, Sphere>();
	private Map<Vertex, TransformGroup> nodeToTG = new HashMap<Vertex, TransformGroup>();
	private Map<Edge, Cylinder> edgeToCylinder = new HashMap<Edge, Cylinder>();
	private Map<Edge, BranchGroup> edgeToBG = new HashMap<Edge, BranchGroup>();
	
	
	// constants for drawing
	final float edgeThickness = .3f;

	final float axisRadius = 0.2f;
	final float axisConeLength = 1.0f;
	final float axisConeRadius = 0.5f;
	final float xAxisLength = 100f;
	final float axisLength = 50f;
	final float axisTranslation = 10f;

//	ColoringAttributes nodeSphereColoringAttributes = new ColoringAttributes();
//	ColoringAttributes edgeColoringAttributes = new ColoringAttributes();
	ColoringAttributes axisColoringAttributes = new ColoringAttributes();

	Material axisMaterial;
//	Material edgeMaterial;


	// Set up colors
	Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
	Color3f white = new Color3f(1.0f, 1.0f, 1.0f);
	Color3f red = new Color3f(1.0f, .0f, .0f);
	Color3f lightred = new Color3f(1.0f, .85f, .8f);
	Color3f blue = new Color3f(0.0f, 0.0f, 1.0f);

	Color3f lightblue = new Color3f(0.25f, 0.25f, .5f);

	Color3f darkblue = new Color3f(.0f, 0.0f, .3f);

	Color3f yellow = new Color3f(1.0f, 1.0f, .0f);

	Color3f darkyellow = new Color3f(.6f, .6f, .0f);

	Color3f lightyellow = new Color3f(0.9f, 0.9f, .65f);

	Color3f orange = new Color3f(1.0f, 0.5f, .0f);

	Color3f green = new Color3f(0.0f, 1.0f, .0f);

	Color3f darkgreen = new Color3f(0.0f, 0.5f, .0f);

	Color3f lightgreen = new Color3f(0.8f, 1.0f, .8f);

	Color3f brown = new Color3f(0.9f, 0.6f, .0f);

	Color3f darkgray = new Color3f(0.3f, 0.3f, .3f);

	Color3f lightgray = new Color3f(0.7f, 0.7f, .7f);

	Color3f darkergray = new Color3f(0.2f, 0.2f, .2f);

	
//	Material nodeSphereMaterial = new Material(black, darkyellow, white, black, 16.0f);


	
	/**
	 * initializes the Graph
	 * 
	 * @param aGraph
	 */
	public BFSVisualizer(DirectedGraph aGraph){
		theGraph = aGraph;
		
		for (Iterator vIt = theGraph.getVertices().iterator(); vIt.hasNext();) {
			Vertex currentNode = (Vertex) vIt.next();
			float currentNodeX = Float.parseFloat((String) currentNode.getUserDatum("x"))/10.0f;
			float currentNodeY = Float.parseFloat((String) currentNode.getUserDatum("y"))/10.0f;
			float currentNodeZ = Float.parseFloat((String) currentNode.getUserDatum("z"))/10.0f;
			nodeToLocation.put(currentNode, new Vector3f(currentNodeX, currentNodeY, currentNodeZ));
		}

	}
	
	/**
	 * Constructs the scene.
	 */
	public void showGraph() {

//		nodeSphereColoringAttributes.setColor(darkyellow);
//		edgeColoringAttributes.setColor(darkgray);
		axisColoringAttributes.setColor(darkgray);
		
//		// ambient, diffuse, specular, emisssive, shininess
//		// edgeMaterial = new Material(black, darkgreen, white, black, 32.0f);
//		edgeMaterial = new Material(black, darkgreen, lightgray, black, 4.0f);
		axisMaterial = new Material(black, darkgray, white, black, 4.0f);
		

		// Create the root of the branch graph
		// sceneRoot is the highest root
		sceneRoot =  new BranchGroup();
		dynamicRoot = new BranchGroup(); 
		dynamicRoot.setCapability(Group.ALLOW_CHILDREN_READ);
		dynamicRoot.setCapability(Group.ALLOW_CHILDREN_WRITE);
		dynamicRoot.setCapability(Group.ALLOW_CHILDREN_EXTEND);
		staticRoot = new BranchGroup(); 
		dynLightRoot = new BranchGroup(); 
		nodeRoot = new BranchGroup(); 
		edgeRoot = new BranchGroup(); 
		dynLightRoot.setCapability(BranchGroup.ALLOW_DETACH);
		edgeRoot.setCapability(BranchGroup.ALLOW_DETACH);
		edgeRoot.setCapability(Group.ALLOW_CHILDREN_WRITE);
		edgeRoot.setCapability(Group.ALLOW_CHILDREN_EXTEND);
		

		sceneRoot.addChild(dynamicRoot);
		sceneRoot.addChild(staticRoot);
		dynamicRoot.addChild(nodeRoot);
		dynamicRoot.addChild(edgeRoot);
		dynamicRoot.addChild(dynLightRoot);
		
		// Set up the lights
		createLight(staticRoot);
		// set up Coordinate System
		createCoordinateSystem(staticRoot);
		// set up graph
		imprintSourceGraph(nodeRoot, edgeRoot);
	}


	/**
	 * Draws nodes, edges and light
	 * 
	 * @param nRoot The Branchgroup for all the nodes.
	 * @param eRoot The Branchgroup for all the edges.
	 */
	private void imprintSourceGraph(BranchGroup nRoot, BranchGroup eRoot) {
		for (Iterator vIt = theGraph.getVertices().iterator(); vIt.hasNext();) {
			Vertex currentNode = (Vertex) vIt.next();

//			Appearance nodeSphereAppearance = new Appearance();
//			nodeSphereColoringAttributes.setShadeModel(ColoringAttributes.NICEST);
//			nodeSphereAppearance.setColoringAttributes(nodeSphereColoringAttributes);
//			nodeSphereAppearance.setMaterial(nodeSphereMaterial);
//			nodeSphereAppearance.setTextureAttributes(nodeTexAttr);
//			nodeSphereAppearance.setTexCoordGeneration(nodeSphereTextCoord);

			float currentNodeRadius = 1;
			Vector3f nodeLocation = getNodeLocation(currentNode);
			
			// create the nodeSphere and attach it
			Sphere currentNodeSphere = new Sphere(currentNodeRadius);
			TransformGroup currentNodeTG = new TransformGroup();
			Transform3D currentNodeTrans3D = new Transform3D();
			Vector3f currentNodeTranslation = new Vector3f(nodeLocation.x, nodeLocation.y, nodeLocation.z);
			currentNodeTrans3D.setTranslation(currentNodeTranslation);
//			currentNodeSphere.setCapability(Sphere.ENABLE_APPEARANCE_MODIFY);
			currentNodeTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

			currentNodeTG.setTransform(currentNodeTrans3D);
			currentNodeTG.addChild(currentNodeSphere);
			
			nodeToSphere.put(currentNode, currentNodeSphere);
			nodeToTG.put(currentNode, currentNodeTG);
			nRoot.addChild(currentNodeTG);
			
		}

		for (Iterator eIt = theGraph.getEdges().iterator(); eIt.hasNext();) {
			Edge currentEdge = (Edge) eIt.next();
			
			eRoot.addChild(placeEdge(currentEdge, edgeThickness, (Vertex) currentEdge.getEndpoints().getFirst(), (Vertex) currentEdge.getEndpoints().getSecond()));

		}
		
	}

//private float spaceOnEdge(Edge edge) {
//	float space = 0.0f;
//	Vector3f sourceLoc = nodeToLocation.get(edge.source());
//	Vector3f targetLoc = nodeToLocation.get(edge.target());
//	Vector3f v = new Vector3f(-sourceLoc.x + targetLoc.x, -sourceLoc.y + targetLoc.y, -sourceLoc.z + targetLoc.z);
//	space = v.length();
//	space -= nodeToSphere.get(edge.source()).getRadius();
//	space -= nodeToSphere.get(edge.source()).getRadius();
//	return (space);
//}
//

private Vector3f getNodeLocation(Vertex currentNode){
	return (Vector3f) nodeToLocation.get(currentNode);
	
}
	
private Sphere getNodeSphere(Vertex currentNode){
	return (Sphere) nodeToSphere.get(currentNode);
	
}
	
/**
 * computes the transformation for placing an edge between two vertices
 * 
 * @param source First node.
 * @param target Second node.
 * @return The transformation.
 */
protected Transform3D getEdgeT3D(Vertex source, Vertex target){
	Vector3f sourceLoc = getNodeLocation(source);
	Vector3f targetLoc = getNodeLocation(target);
	Vector3f v = new Vector3f(-sourceLoc.x + targetLoc.x, -sourceLoc.y + targetLoc.y, -sourceLoc.z + targetLoc.z);
	
	Transform3D T3D = new Transform3D();
	Vector3f Translation = new Vector3f(0.5f * (sourceLoc.x + targetLoc.x), 0.5f * (sourceLoc.y + targetLoc.y), 0.5f * (sourceLoc.z + targetLoc.z));

	Vector3f RotAxis = new Vector3f(0.0f, 1.0f, 0.0f);
	float RotAngle = (float) Math.acos((RotAxis.dot(v) / (v.length() * RotAxis.length())));
	RotAxis.cross(RotAxis, v);
	AxisAngle4f Rotation = new AxisAngle4f(RotAxis.x, RotAxis.y, RotAxis.z, RotAngle);
	T3D.setTranslation(Translation);
	T3D.setRotation(Rotation);
	
	
	return T3D;
}



/**
 * Draws an edge.
 * 
 * @param theEdge
 * @param thickness
 * @param source
 * @param target
 * @return The BranchGroup of the created edge.
 */
private javax.media.j3d.Node placeEdge(Edge theEdge, float thickness, Vertex source, Vertex target) {
	Vector3f sourceLoc = getNodeLocation(source);
	Vector3f targetLoc = getNodeLocation(target);
	Vector3f direction = new Vector3f(-sourceLoc.x + targetLoc.x, -sourceLoc.y + targetLoc.y, -sourceLoc.z + targetLoc.z);
	float length = (float) direction.length();

//	Appearance edgeAppearance = new Appearance();
//	edgeColoringAttributes.setShadeModel(ColoringAttributes.NICEST);
//	edgeAppearance.setColoringAttributes(edgeColoringAttributes);
//	edgeAppearance.setMaterial(edgeMaterial);

	// the edge
	Cylinder cylinder = new Cylinder(thickness, length);
	
	BranchGroup bg = new BranchGroup();
	bg.setCapability(BranchGroup.ALLOW_DETACH);
	TransformGroup TG = new TransformGroup();
	bg.addChild(TG);
	Transform3D T3D = getEdgeT3D(source, target);
	TG.setTransform(T3D);

	{ // Pfeilspitze bauen und auch ausrichten
		// if (restrictArrowCones && (spaceOnEdge(theEdge) >= thickness * 6.0f)) {

		Cone edgeCone = new Cone(thickness * 3.0f, thickness * 6.0f);//, edgeAppearance);
		TransformGroup TGCone = new TransformGroup();	
		Transform3D T3DCone = new Transform3D();
		Vector3f translationOfCone = new Vector3f(direction);
		translationOfCone.scale(0.5f);
		T3DCone.setTranslation(translationOfCone);
		// TGCone.setTransform(T3DCone);
		TGCone.addChild(edgeCone);
		TG.addChild(TGCone);
		
		// add the cone to Map
//		edgeToCone.put(theEdge, edgeCone);
	}

	TG.addChild(cylinder);
	// add the cylinder to Map
	TG.setPickable(false);
	edgeToCylinder.put(theEdge, cylinder);
	edgeToBG.put(theEdge, bg);
	return bg;
}

	
protected void deleteEdge(Edge e){
	// TODO: this only detaches the edge instead of really deleting it.
	BranchGroup bg = (BranchGroup) edgeToBG.get(e);
	bg.detach();
	
}


	
	/**
	 * Draws the coordinate system.
	 * 
	 * @param objRoot The sub-scene is attached to this BranchGroup.
	 */
	private void createCoordinateSystem(BranchGroup objRoot) {

		Appearance axisAppearance = new Appearance();
		axisAppearance.setMaterial(axisMaterial);
		axisColoringAttributes.setShadeModel(ColoringAttributes.NICEST);
		axisAppearance.setColoringAttributes(axisColoringAttributes);

		// set up origin
		Sphere theOrigin = new Sphere(axisRadius * 1.5f);
		TransformGroup originTG = new TransformGroup();
		Transform3D originTrans3D = new Transform3D();
		Vector3f originPosition = new Vector3f(-1f * xAxisLength / 2f, -1f * axisTranslation, -1f * axisTranslation);
		originTrans3D.setTranslation(originPosition);
		originTG.setTransform(originTrans3D);
		originTG.addChild(theOrigin);
		objRoot.addChild(originTG);

		// set up axes
		Cylinder xAxis = new Cylinder(axisRadius, xAxisLength, axisAppearance);
		Cylinder yAxis = new Cylinder(axisRadius, axisLength, axisAppearance);
		Cylinder zAxis = new Cylinder(axisRadius, axisLength, axisAppearance);
		TransformGroup xAxisTG = new TransformGroup();
		TransformGroup yAxisTG = new TransformGroup();
		TransformGroup zAxisTG = new TransformGroup();
		Transform3D xAxisTrans3D = new Transform3D();
		Transform3D yAxisTrans3D = new Transform3D();
		Transform3D zAxisTrans3D = new Transform3D();
		Vector3f xAxisPosition = new Vector3f(0.0f, -1f * axisTranslation, -1f * axisTranslation);
		Vector3f yAxisPosition = new Vector3f(-1f * xAxisLength / 2f, (float) (axisLength / 2.0 - axisTranslation), -1f * axisTranslation);
		Vector3f zAxisPosition = new Vector3f(-1f * xAxisLength / 2f, -1f * axisTranslation, (float) (axisLength / 2.0 - axisTranslation));
		xAxisTrans3D.setTranslation(xAxisPosition);
		yAxisTrans3D.setTranslation(yAxisPosition);
		zAxisTrans3D.setTranslation(zAxisPosition);
		xAxisTrans3D.setRotation(new AxisAngle4d(0, 0, 1, Math.PI / 2.0));
		yAxisTrans3D.setRotation(new AxisAngle4d(0, 0, 0, 0)); // no rotation
		zAxisTrans3D.setRotation(new AxisAngle4d(1, 0, 0, Math.PI / 2.0));
		xAxisTG.setTransform(xAxisTrans3D);
		yAxisTG.setTransform(yAxisTrans3D);
		zAxisTG.setTransform(zAxisTrans3D);
		xAxisTG.addChild(xAxis);
		yAxisTG.addChild(yAxis);
		zAxisTG.addChild(zAxis);
		objRoot.addChild(xAxisTG);
		objRoot.addChild(yAxisTG);
		objRoot.addChild(zAxisTG);

		Cone xAxisCone = new Cone(axisConeRadius, axisConeLength, axisAppearance);
		Cone yAxisCone = new Cone(axisConeRadius, axisConeLength, axisAppearance);
		Cone zAxisCone = new Cone(axisConeRadius, axisConeLength, axisAppearance);
		TransformGroup xAxisConeTG = new TransformGroup();
		TransformGroup yAxisConeTG = new TransformGroup();
		TransformGroup zAxisConeTG = new TransformGroup();
		Transform3D xAxisConeTrans3D = new Transform3D();
		Transform3D yAxisConeTrans3D = new Transform3D();
		Transform3D zAxisConeTrans3D = new Transform3D();
		Vector3f xAxisConePosition = new Vector3f(xAxisLength / 2.0f + axisConeLength / 2.0f, -1f * axisTranslation, -1f * axisTranslation);
		Vector3f yAxisConePosition = new Vector3f(-1f * xAxisLength / 2f, (float) (axisLength - axisTranslation) + axisConeLength / 2.0f, -1f * axisTranslation);
		Vector3f zAxisConePosition = new Vector3f(-1f * xAxisLength / 2f, -1f * axisTranslation, (float) (axisLength - axisTranslation) + axisConeLength / 2.0f);
		xAxisConeTrans3D.setTranslation(xAxisConePosition);
		yAxisConeTrans3D.setTranslation(yAxisConePosition);
		zAxisConeTrans3D.setTranslation(zAxisConePosition);
		xAxisConeTrans3D.setRotation(new AxisAngle4d(0, 0, -1, Math.PI / 2.0));
		yAxisConeTrans3D.setRotation(new AxisAngle4d(0, 0, 0, 0)); // no
		// rotation
		zAxisConeTrans3D.setRotation(new AxisAngle4d(1, 0, 0, Math.PI / 2.0));
		xAxisConeTG.setTransform(xAxisConeTrans3D);
		yAxisConeTG.setTransform(yAxisConeTrans3D);
		zAxisConeTG.setTransform(zAxisConeTrans3D);
		xAxisConeTG.addChild(xAxisCone);
		yAxisConeTG.addChild(yAxisCone);
		zAxisConeTG.addChild(zAxisCone);
		
		// ... finally! 
		objRoot.addChild(xAxisConeTG);
		objRoot.addChild(yAxisConeTG);
		objRoot.addChild(zAxisConeTG);
	}

	
	/**
	 * Let there be light.
	 * 
	 * @param objRoot The sub-scene is attached to this BranchGroup.
	 */
	private void createLight(BranchGroup objRoot) {
		BoundingSphere lightBounds = new BoundingSphere(new Point3d(10.0, 10.0, 10.0), 200.0);

//		// Set up the ambient light
//		Color3f ambientColor = new Color3f(1.0f, 1.0f, 1.0f);
//		AmbientLight ambientLightNode = new AmbientLight(ambientColor);
//		ambientLightNode.setInfluencingBounds(lightBounds);
//		// objRoot.addChild(ambientLightNode);

		/* Set up the directional lights */
		Color3f light1Color = orange;
		Vector3f light1Direction = new Vector3f(-1.0f, -1.0f, -1.0f);
		DirectionalLight light1 = new DirectionalLight(light1Color, light1Direction);
		light1.setInfluencingBounds(lightBounds);
		objRoot.addChild(light1);

		Color3f light2Color = yellow;
		Vector3f light2Direction = new Vector3f(1.0f, 0.2f, 0.2f);
		DirectionalLight light2 = new DirectionalLight(light2Color, light2Direction);
		light2.setInfluencingBounds(lightBounds);
		objRoot.addChild(light2);

		Color3f light3Color = red;
		Vector3f light3Direction = new Vector3f(0.2f, 1.0f, 0.2f);
		DirectionalLight light3 = new DirectionalLight(light3Color, light3Direction);
		light3.setInfluencingBounds(lightBounds);
		objRoot.addChild(light3);

		Color3f light4Color = blue;
		Vector3f light4Direction = new Vector3f(0.2f, 0.2f, 1.0f);
		DirectionalLight light4 = new DirectionalLight(light4Color, light4Direction);
		light4.setInfluencingBounds(lightBounds);
		objRoot.addChild(light4);

	}

	
	public BranchGroup getChild() {
		return sceneRoot;
	}

	
	public Vector3d getStartPosition() {
		Vector3d result = new Vector3d(0,0,50);
//		if (theGraph != null){
//			Vector3f sum = new Vector3f(0,0,0);
//			for(Iterator vertices = theGraph.getVertices().iterator(); vertices.hasNext();){
//				sum.add(nodeToLocation.get(vertices.next()));
//			}
//			int n = theGraph.numVertices();
//			result.add(new Vector3d(sum.x / n, sum.y / n, sum.z /n));
//		}
	
		// put the camera at the center of the graph and far enough to see all vertices
		if (theGraph != null){
			double posX=0.0, posY=0.0, posZ=0.0;
			double xRange = 0.0, yRange = 0.0;
			double xMin = nodeToLocation.get(theGraph.getVertices().iterator().next()).x;
			double xMax = nodeToLocation.get(theGraph.getVertices().iterator().next()).x;
			double yMin = nodeToLocation.get(theGraph.getVertices().iterator().next()).y;
			double yMax = nodeToLocation.get(theGraph.getVertices().iterator().next()).y;
			double zMax = nodeToLocation.get(theGraph.getVertices().iterator().next()).z;
			Vertex currentVertex = null;
			for(Iterator vertices = theGraph.getVertices().iterator(); vertices.hasNext();){
				currentVertex = (Vertex) vertices.next();
				if (nodeToLocation.get(currentVertex).x < xMin) xMin = nodeToLocation.get(currentVertex).x;
				if (nodeToLocation.get(currentVertex).x > xMax) xMax = nodeToLocation.get(currentVertex).x;
				if (nodeToLocation.get(currentVertex).y < yMin) yMin = nodeToLocation.get(currentVertex).y;
				if (nodeToLocation.get(currentVertex).y > yMax) yMax = nodeToLocation.get(currentVertex).y;
				if (nodeToLocation.get(currentVertex).z > zMax) zMax = nodeToLocation.get(currentVertex).z;
			}
			xRange = xMax - xMin;
			yRange = yMax - yMin;
			posX = xMin + xRange/2.0;
			posY = yMin + yRange/2.0;
			// the camera angle is assumed to be 45 degrees
			posZ = zMax + Math.max(xRange/2.0,yRange/2.0) / Math.tan(Math.PI / 8.0);
			result = (new Vector3d(posX, posY, posZ));
		}
		return  new Vector3d(result.x, result.y, result.z);
	}
	
	/**
	 * Places a Vertex at a specified location. 
	 * 
	 * @param v	The vertex.
	 * @param loc The new location.
	 */
	public void setNodeLocation(Vertex v, Vector3f loc){
		Sphere sphere =  getNodeSphere(v);
		
		TransformGroup tg = (TransformGroup) nodeToTG.get(v);
		Transform3D t3d = new Transform3D();
		t3d.setTranslation(loc);
		tg.setTransform(t3d);
		
		for (Iterator adjEdges = v.getIncidentEdges().iterator()
				; adjEdges.hasNext();){
			Edge currentEdge = (Edge) adjEdges.next();
			
			deleteEdge(currentEdge);
			edgeRoot.addChild(placeEdge(currentEdge, edgeThickness, (Vertex) currentEdge.getEndpoints().getFirst(), (Vertex) currentEdge.getEndpoints().getSecond()));
		}
	}

	
	/**
	 * A simple way of highlighting a node.
	 * 
	 * @param v
	 */
	public void highlight(Vertex v){
		Vector3f loc= getNodeLocation(v);
		loc.z += 10f;
		setNodeLocation(v, loc);
		
	}
}
