/*
 * Created on 04.01.2006
 *
 * Copyright --- All rights reserved.
 * ITI, Universität Karlsruhe, Germany.
 */
package de.uka.algo.GT3D;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import javax.vecmath.Point3d;

import de.uka.algo.GT3D.exceptions.BadStateException;
import de.uka.algo.GT3D.exceptions.FileFormatException;
import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.impl.DirectedSparseGraph;
import edu.uci.ics.jung.io.GraphMLFile;

/**
 * 
 * This class implements Breadth First Search.
 * @author stm
 *
 */
public class BreadthFirstSearchDemo implements StepwiseAlgorithm {

	
	/**
	 * The graph we are dealing with, Read in <code>init</code>. 
	 */
	private DirectedGraph theGraph = null;
	/**
	 * Waiting list for nodes
	 */
	LinkedList<Vertex> candidates = null;
	/**
	 * Map of visited nodes
	 */
	Map<Vertex, Boolean> visited = null;

	boolean initialized = false;
	boolean ready = false;
	
	
	public Point3d getBoundingBox() {
		return null;
	}

	
	public void init(DirectedGraph aGraph){
		theGraph = aGraph;
		init();
	}
	
	public void init(){
		if (theGraph == null)
			throw new BadStateException("Graph not initialized.");
			
		
		
		/* initialize data structures to store the state of the search */
		candidates = new LinkedList<Vertex>();
		visited = new HashMap<Vertex, Boolean>();
		
		if (theGraph == null) {
			ready = true;
		}else{
			// initialize map of visited nodes
			Vertex e = null;
			for (Iterator nodes=theGraph.getVertices().iterator()
					;nodes.hasNext();){
				e = (Vertex) nodes.next();
				visited.put(e, new Boolean(false));
			}
			if (e == null){
				// empty graph -- we are done
				ready = true;
			}else{
				// put first node into waiting list
				candidates.addLast(e);
				visited.put(e, true);
			}
			
		}
		
		initialized = true;
		
	}
	/**
	 * Initialize the graph, the map of visited nodes and put the first 
	 * vertex into the queue.
	 * 
	 * 
	 * @param filename the name of the GraphML file to read from.
	 * 
	 * @see de.uka.algo.GT3D.StepwiseAlgorithm#init(java.lang.String)
	 */
	public void init(String filename) throws FileFormatException {
		// create a GraphML file reader which is appropriate for the specified
		// format and load the file specified by filename
		GraphMLFile file = new GraphMLFile(new GT3DGraphMLFileHandler());
		Graph aGraph = file.load(filename);
		if (! (aGraph instanceof DirectedSparseGraph)) {
			throw new FileFormatException();
		}
		theGraph = (DirectedGraph) aGraph;

		init();
	}

	
	/** 
	 * Perform the next step of the search:
	 * Get the next vertex fromt the queue and mark. all its neighbours.
	 *
	 * @see de.uka.algo.GT3D.StepwiseAlgorithm#nextStep()
	 */
	public void nextStep() {
		if (!initialized)
			throw new BadStateException("Graph must be initilized befor it is run.");

		// get first vertex from queue
		Vertex currentVertex = (Vertex) candidates.removeFirst();
		if (currentVertex == null){
			throw new RuntimeException("unexpected error");
		}
		
		// add all unvisited neighbours to queue
		for (Iterator neighbours = currentVertex.getNeighbors().iterator()
				; neighbours.hasNext();){
			Vertex currentNeighbour = (Vertex) neighbours.next();
			if (! (Boolean) visited.get(currentNeighbour)){
				candidates.addLast(currentNeighbour);
				// mark current vertex as visited
				visited.put(currentNeighbour, true);
			}
		}
		
		// do visualization
		markNode(currentVertex);
			
		// search for next unvisited connected component if this one is
		// completely visited
		if (candidates.isEmpty()){
			for (Iterator nodes=theGraph.getVertices().iterator();nodes.hasNext();){
				Vertex v = (Vertex) nodes.next();
				if ((Boolean) visited.get(v) == false){
					candidates.addLast(v);
					return;
				}
			}
			// no unvisited nodes? We are done.
			ready = true;
		}
	}

	/**
	 * Visualize a visited node. For now we only print the node.
	 * 
	 * @param v the current node.
	 */
	protected void markNode(Vertex v){
		System.out.println(v);
	}
	
	/**
	 * this method does nothing
	 * 
	 * @see de.uka.algo.GT3D.StepwiseAlgorithm#previousStep()
	 */
	public void previousStep() {
		// for now we do not do anything
		return;
	}

	public boolean finished() {
		return ready;
	}



	public boolean atStart() {
		return false;
	}

}
